@@ -7,45 +7,48 @@ use std::net::ToSocketAddrs;
 use std::sync::Arc;
 use std::thread::{sleep, spawn};
 use std::time::Duration;
-use std::collections::{BTreeMap, VecDeque};
+use std::collections::VecDeque;
 
 #[derive(Clone, Debug, Default, Eq, PartialEq, Serialize)]
 struct StatusView {
     done: bool,
     model: String,
     generated: usize,
-    discoveries: BTreeMap<String, String>, // name+classification -> encoded path
+    properties: Vec<(Expectation,
+                     String,           // name
+                     Option<String>)>, // encoded path to discovery
     recent_path: Option<String>,
 }
 
 #[derive(Debug, Eq, PartialEq)]
-struct StateView<State, Action> {
-    action: Option<Action>,
+struct StateView<State> {
+    action: Option<String>,
     outcome: Option<String>,
-    state: State,
+    state: Option<State>,
     svg: Option<String>,
 }
 
-type StateViewsJson<State, Action> = Json<Vec<StateView<State, Action>>>;
+type StateViewsJson<State> = Json<Vec<StateView<State>>>;
 
-impl<Action, State> serde::Serialize for StateView<State, Action>
+impl<State> serde::Serialize for StateView<State>
 where
-    Action: Debug,
     State: Debug + Hash,
 {
     fn serialize<S: Serializer>(&self, ser: S) -> Result<S::Ok, S::Error> {
         let mut out = ser.serialize_struct("StateView", 3)?;
         if let Some(ref action) = self.action {
-            out.serialize_field("action", &format!("{:?}", action))?;
+            out.serialize_field("action", action)?;
         }
         if let Some(ref outcome) = self.outcome {
             out.serialize_field("outcome", outcome)?;
         }
-        out.serialize_field("state", &format!("{:#?}", self.state))?;
+        if let Some(ref state) = self.state {
+            out.serialize_field("state", &format!("{:#?}", state))?;
+            out.serialize_field("fingerprint", &format!("{:?}", fingerprint(&state)))?;
+        }
         if let Some(ref svg) = self.svg {
             out.serialize_field("svg", svg)?;
         }
-        out.serialize_field("fingerprint", &format!("{:?}", fingerprint(&self.state)))?;
         out.end()
     }
 }
@@ -139,12 +142,12 @@ where M: Model,
         model: std::any::type_name::<M>().to_string(),
         done: checker.is_done(),
         generated: checker.generated_count(),
-        discoveries: checker.discoveries().into_iter()
-            .map(|(name, path)| {
-                let key = format!("\"{}\" {}", name, checker.discovery_classification(name));
-                let value = path.encode();
-                (key, value)
-            })
+        properties: checker.model().properties().into_iter()
+            .map(|p| (
+                    p.expectation,
+                    p.name.to_string(),
+                    checker.discovery(p.name).map(|p| p.encode()),
+            ))
             .collect(),
         recent_path: snapshot.read().1.as_ref().map(|p| format!("{:?}", p)),
     };
@@ -152,8 +155,9 @@ where M: Model,
 }
 
 fn states<M, C>(req: HttpRequest, data: Data<M::Action, C>)
-    -> Result<StateViewsJson<M::State, M::Action>>
+    -> Result<StateViewsJson<M::State>>
 where M: Model,
+      M::Action: Debug,
       M::State: Debug + Hash,
       C: Checker<M>,
 {
@@ -186,7 +190,7 @@ where M: Model,
             results.push(StateView {
                 action: None,
                 outcome: None,
-                state,
+                state: Some(state),
                 svg,
             });
         }
@@ -200,7 +204,7 @@ where M: Model,
         model.actions(&last_state, &mut actions2);
         model.actions(&last_state, &mut actions3);
         for ((action, action2), action3) in actions1.into_iter().zip(actions2).zip(actions3) {
-            let outcome = model.display_outcome(&last_state, action2);
+            let outcome = model.format_step(&last_state, action2);
             let state = model.next_state(&last_state, action3);
             if let Some(state) = state {
                 let svg = {
@@ -208,7 +212,20 @@ where M: Model,
                     fingerprints.push_back(fingerprint(&state));
                     model.as_svg(Path::from_fingerprints::<M>(model, fingerprints))
                 };
-                results.push(StateView { action: Some(action), outcome, state, svg });
+                results.push(StateView {
+                    action: Some(model.format_action(&action)),
+                    outcome,
+                    state: Some(state),
+                    svg,
+                });
+            } else {
+                // "Action ignored" case is still returned, as it may be useful for debugging.
+                results.push(StateView {
+                    action: Some(model.format_action(&action)),
+                    outcome: None,
+                    state: None,
+                    svg: None,
+                });
             }
         }
     } else {
@@ -229,8 +246,8 @@ mod test {
     fn can_init() {
         let checker = Arc::new(BinaryClock.checker().spawn_bfs());
         assert_eq!(get_states(Arc::clone(&checker), "/").unwrap(), vec![
-            StateView { action: None, outcome: None, state: 0, svg: None },
-            StateView { action: None, outcome: None, state: 1, svg: None },
+            StateView { action: None, outcome: None, state: Some(0), svg: None },
+            StateView { action: None, outcome: None, state: Some(1), svg: None },
         ]);
     }
 
@@ -247,9 +264,9 @@ mod test {
         // ```
         assert_eq!(get_states(Arc::clone(&checker), "/2716592049047647680/9080728272894440685").unwrap(), vec![
             StateView {
-                action: Some(BinaryClockAction::GoHigh),
+                action: Some("GoHigh".to_string()),
                 outcome: Some("1".to_string()),
-                state: 1,
+                state: Some(1),
                 svg: None,
             },
         ]);
@@ -268,7 +285,6 @@ mod test {
     fn smoke_test_states() {
         use crate::actor::{ActorModelState, DuplicatingNetwork, Envelope, Id, LossyNetwork};
         use crate::actor::actor_test_util::ping_pong::{PingPongCfg, PingPongMsg::*};
-        use crate::actor::ActorModelAction::*;
         use crate::util::HashableHashSet;
         use std::iter::FromIterator;
 
@@ -288,14 +304,14 @@ mod test {
                 StateView {
                     action: None,
                     outcome: None,
-                    state: ActorModelState {
+                    state: Some(ActorModelState {
                         actor_states: vec![Arc::new(0), Arc::new(0)],
                         history: (0, 1),
                         is_timer_set: vec![],
                         network: HashableHashSet::from_iter(vec![
                             Envelope { src: Id::from(0), dst: Id::from(1), msg: Ping(0) },
                         ]),
-                    },
+                    }),
                     svg: Some("<svg version=\'1.1\' baseProfile=\'full\' width=\'500\' height=\'30\' viewbox=\'-20 -20 520 50\' xmlns=\'http://www.w3.org/2000/svg\'><defs><marker class=\'svg-event-shape\' id=\'arrow\' markerWidth=\'12\' markerHeight=\'10\' refX=\'12\' refY=\'5\' orient=\'auto\'><polygon points=\'0 0, 12 5, 0 10\' /></marker></defs><line x1=\'0\' y1=\'0\' x2=\'0\' y2=\'30\' class=\'svg-actor-timeline\' />\n<text x=\'0\' y=\'0\' class=\'svg-actor-label\'>0</text>\n<line x1=\'100\' y1=\'0\' x2=\'100\' y2=\'30\' class=\'svg-actor-timeline\' />\n<text x=\'100\' y=\'0\' class=\'svg-actor-label\'>1</text>\n</svg>\n".to_string()),
                 },
             ]);
@@ -318,22 +334,22 @@ mod test {
         assert_eq!(
             states[0],
             StateView {
-                action: Some(Drop(Envelope { src: Id::from(0), dst: Id::from(1), msg: Ping(0) })),
+                action: Some("Drop(Envelope { src: Id(0), dst: Id(1), msg: Ping(0) })".to_string()),
                 outcome: Some("DROP: Envelope { src: Id(0), dst: Id(1), msg: Ping(0) }".to_string()),
-                state: ActorModelState {
+                state: Some(ActorModelState {
                     actor_states: vec![Arc::new(0), Arc::new(0)],
                     history: (0, 1),
                     is_timer_set: vec![],
                     network: HashableHashSet::new(),
-                },
+                }),
                 svg: Some("<svg version='1.1' baseProfile='full' width='500' height='60' viewbox='-20 -20 520 80' xmlns='http://www.w3.org/2000/svg'><defs><marker class='svg-event-shape' id='arrow' markerWidth='12' markerHeight='10' refX='12' refY='5' orient='auto'><polygon points='0 0, 12 5, 0 10' /></marker></defs><line x1='0' y1='0' x2='0' y2='60' class='svg-actor-timeline' />\n<text x='0' y='0' class='svg-actor-label'>0</text>\n<line x1='100' y1='0' x2='100' y2='60' class='svg-actor-timeline' />\n<text x='100' y='0' class='svg-actor-label'>1</text>\n</svg>\n".to_string()),
             });
         assert_eq!(
             states[1],
             StateView {
-                action: Some(Deliver { src: Id::from(0), dst: Id::from(1), msg: Ping(0) }),
+                action: Some("Id(0) → Ping(0) → Id(1)".to_string()),
                 outcome: Some("OUT: [Send(Id(0), Pong(0))]\n\nNEXT_STATE: 1\n\nPREV_STATE: 0\n".to_string()),
-                state: ActorModelState {
+                state: Some(ActorModelState {
                     actor_states: vec![
                         Arc::new(0),
                         Arc::new(1),
@@ -343,7 +359,7 @@ mod test {
                     network: HashableHashSet::from_iter(vec![
                         Envelope { src: Id::from(1), dst: Id::from(0), msg: Pong(0) },
                     ]),
-                },
+                }),
                 svg: Some("<svg version='1.1' baseProfile='full' width='500' height='60' viewbox='-20 -20 520 80' xmlns='http://www.w3.org/2000/svg'><defs><marker class='svg-event-shape' id='arrow' markerWidth='12' markerHeight='10' refX='12' refY='5' orient='auto'><polygon points='0 0, 12 5, 0 10' /></marker></defs><line x1='0' y1='0' x2='0' y2='60' class='svg-actor-timeline' />\n<text x='0' y='0' class='svg-actor-label'>0</text>\n<line x1='100' y1='0' x2='100' y2='60' class='svg-actor-timeline' />\n<text x='100' y='0' class='svg-actor-label'>1</text>\n<line x1='0' x2='100' y1='0' y2='30' marker-end='url(#arrow)' class='svg-event-line' />\n<text x='100' y='30' class='svg-event-label'>Ping(0)</text>\n</svg>\n".to_string()),
             });
     }
@@ -371,15 +387,30 @@ mod test {
                  stateright::actor::actor_test_util::ping_pong::PingPongActor, \
                  stateright::actor::actor_test_util::ping_pong::PingPongCfg, (u32, u32)>");
         assert_eq!(status.generated, 5);
-        assert_eq!(status.discoveries.len(), 2);
-        assert!(status.discoveries.get("\"can reach max\" example").is_some());
-        assert!(status.discoveries.get("\"must exceed max\" counterexample").is_some());
+        assert_eq!(status.properties, vec![
+            (Expectation::Always, "delta within 1".to_string(), None),
+            (Expectation::Sometimes, "can reach max".to_string(), Some(
+                    "2298670378538534683\
+                    /6952966106190701737\
+                    /713954962230533912\
+                    /13964153056603498260".to_string())),
+            (Expectation::Eventually, "must reach max".to_string(), None),
+            (Expectation::Eventually, "must exceed max".to_string(), Some(
+                    "2298670378538534683\
+                    /6952966106190701737\
+                    /713954962230533912\
+                    /13964153056603498260\
+                    /8719175338738027539".to_string())),
+            (Expectation::Always, "#in <= #out".to_string(), None),
+            (Expectation::Eventually, "#out <= #in + 1".to_string(), None)
+        ]);
         assert!(status.recent_path.unwrap().starts_with("["));
     }
 
     fn get_states<M, C>(checker: Arc<C>, path_name: &'static str)
-                -> Result<Vec<StateView<M::State, M::Action>>>
+                -> Result<Vec<StateView<M::State>>>
     where M: Model,
+          M::Action: Debug,
           M::State: Debug + Hash,
           C: Checker<M>,
     {
